#!/usr/bin/:env python
# coding: utf-8

import torch
import matplotlib.pyplot as plt
from PIL import Image
from torchvision import transforms
import numpy as np
from scipy.cluster.vq import kmeans
import torch.nn.functional as F
from datetime import datetime
import os
from pdb import set_trace

## device
device = 'cuda' if torch.cuda.is_available() else 'cpu'
# device = 'cpu'
print('device=', device)
torch.set_default_dtype(torch.float)

## logs
if not os.path.exists('results'):
    os.mkdir('results')

## read image
img_ = np.array(Image.open('butterfly.png')) / 255.
ss_fac = 1    # subsampling factor
if ss_fac is not 1:
    img_ = img_[np.arange(0, img_.shape[0], ss_fac), :, :][:, np.arange(0, img_.shape[1], ss_fac), :]
img = torch.tensor(img_, device=device)
print('img.type=', img.type)
ny, nx, nc = img.shape
print('nx, ny, nc =', nx, ny, nc)


## compute unary energy f via k-means
K = 4
centroids, _ = kmeans(np.reshape(img_, [nx*ny, nc]), K)
print('centroids.shape=', centroids.shape)
f_ = np.sqrt(np.sum((img_[:,:,:,np.newaxis] - np.transpose(centroids)[np.newaxis,np.newaxis,:,:]) ** 2, axis=2))
print('f.shape=', f_.shape)
f = torch.tensor(f_, device=device)


## hyperparameters
normalize_mesh = True
h = np.sqrt(nx * ny) if normalize_mesh is True else 1    # inverse mesh size
print('inverse_meshsize=', h)
alpha = 2e-3    # TV regularization weight
gamma = 0e-1     # Huber parameter (set to 0 if no Huber smoothing)
normK = torch.sqrt(torch.tensor(8.)) * h
max_iters = 500
tol = 1e-2
s = 1.001e-2 * normK; t = 1.001e2 * normK


## initialize optim variables
# u = F.softmax(torch.randn(ny, nx, K, device=device), dim=2)    # random init
u = F.softmax(torch.ones(ny, nx, K, device=device), dim=2)    # uniform init
p = [torch.zeros(ny, nx-1, K, device=device), torch.zeros(ny-1, nx, K, device=device)]


## gradient operator and its adjoint
def Nabla(u, h=1, device='cpu'):
    ny, nx, K = u.shape
    px = u[:, 1:, :] - u[:, :-1, :]
    py = u[1:, :, :] - u[:-1, :, :]
    return [px, py] if h==1 else [px * h, py * h]

def NablaT(p, h=1, device='cpu'):
    ny, nx, K = p[0].shape; nx = nx + 1
    ux = torch.zeros(ny, nx, K, device=device)
    ux[:, 0, :] = -p[0][:, 0, :]; 
    ux[:, -1, :] = p[0][:, -1, :]
    ux[:, 1:-1, :] = p[0][:, :-1, :] - p[0][:, 1:, :]
    uy = torch.zeros(ny, nx, K, device=device)
    uy[0, :, :] = -p[1][0, :, :]; uy[-1, :, :] = p[1][-1, :, :]
    uy[1:-1, :, :] = p[1][:-1, :, :] - p[1][1:, :, :]
    return ux + uy if h==1 else (ux + uy) * h


## projection onto probability simplex [Duchi et al. ICML'08]
def proj_simplex(v, z=1, device='cpu'):
    n, K = v.shape
    u = torch.sort(v, dim=1, descending=True)[0]
    u_cumsum = torch.cumsum(u, dim=1) - z
    idx = torch.arange(K, device=device)
    u_cumsum /= (idx + 1)
    is_gt = torch.gt(u, u_cumsum)
    rho = torch.max(is_gt * idx, dim=1, keepdim=False)[0]
    theta = u_cumsum[torch.arange(n, device=device), rho].view(n, 1)
    out = torch.clamp_(v - theta, min=0)
    return out


## PDHG iterations
primal_energy = []
dual_energy = []
duality_gap = []
run_time = []
nabla_u = Nabla(u, h, device=device)
timer = datetime.now()
print('PDHG starts...')
for it in range(max_iters):
    ## primal update
    u_tilde = u - (NablaT(p, h, device=device) + f) / s
    u = proj_simplex(u_tilde.view(-1, K), device=device).view(ny, nx, K)
    nabla_u_prev = nabla_u
    nabla_u = Nabla(u, h, device=device)
    ## dual update
    p[0] = torch.clamp_((p[0] * t + (nabla_u[0] * 2 - nabla_u_prev[0])) / (t + gamma), min=-alpha, max=alpha)
    p[1] = torch.clamp_((p[1] * t + (nabla_u[1] * 2 - nabla_u_prev[1])) / (t + gamma), min=-alpha, max=alpha)
    ## validation
    if gamma>0:  # Huber TV
        idx = [torch.gt(nabla_u[0].abs(), gamma).float(), torch.gt(nabla_u[1].abs(), gamma).float()]
        primal_energy.append( ((f*u).sum() + (((nabla_u[0].abs() - gamma/2) * idx[0] + (nabla_u[0].pow(2) /gamma/2) * (1-idx[0])).sum() + ((nabla_u[1].abs() - gamma/2) * idx[1] + (nabla_u[1].pow(2) /gamma/2) * (1-idx[1])).sum()) * alpha) / (h**2) )
        dual_energy.append( (torch.sum(torch.min(NablaT(p, h, device=device) + f, dim=-1)[0]) + (p[0].pow(2).sum() + p[1].pow(2).sum()) * gamma/2) / (h**2) )
    else:        # original TV
        primal_energy.append( ((f*u).sum() + (nabla_u[0].abs().sum() + nabla_u[1].abs().sum()) * alpha) / (h**2) )
        dual_energy.append( torch.sum(torch.min(NablaT(p, h, device=device) + f, -1)[0]) / (h**2))
    duality_gap.append( primal_energy[-1] - dual_energy[-1] )
    run_time.append( (datetime.now()-timer).total_seconds() )
    ## print stats
    if (it+1) % 10 == 0:
        print('it {}: primal-ener= {:.6g}, dual-ener= {:.6g}, duality-gap= {:.4e}'.format(
            it+1, primal_energy[-1], dual_energy[-1], duality_gap[-1]))
    ## stopping rule
    if duality_gap[-1] / duality_gap[0] < tol:
        print('PDHG terminated at iter {} with desired accuracy.'.format(it+1))
        break
    elif it ==  max_iters - 1:
        print('PDHG terminated after maximum iterations.')
print('Time elapsed: {:.2f} seconds.'.format(run_time[-1]))


## visualize result
plt.figure()
plt.subplot(121)
plt.plot(primal_energy)
plt.title('primal energy')
plt.subplot(122)
plt.semilogy(duality_gap)
plt.title('duality gap')
plt.savefig('./results/stats.pdf', format='pdf')

classes = torch.max(u.view(-1, K), dim=-1)[1].cpu().numpy()
img_segm = centroids[classes, :].reshape([ny, nx, 3])
plt.figure()
plt.subplot(121)
plt.imshow(img_)
plt.title('image')
plt.subplot(122)
plt.imshow(img_segm)
plt.title('segmentation')
plt.savefig('./results/segmentation.pdf', format='pdf')

u_ = u.cpu().numpy()
umax = np.max(u_)
plt.figure(figsize=(10, 8))
plt.subplot(221)
plt.imshow(u_[:, :, 0], vmin=0, vmax=umax)
plt.title('u[:,:,0]')
plt.subplot(222)
plt.imshow(u_[:, :, 1], vmin=0, vmax=umax)
plt.title('u[:,:,1]')
plt.subplot(223)
plt.imshow(u_[:, :, 2], vmin=0, vmax=umax)
plt.title('u[:,:,2]')
plt.subplot(224)
plt.imshow(u_[:, :, 3], vmin=0, vmax=umax)
plt.title('u[:,:,3]')
plt.savefig('./results/labels.pdf', format='pdf')


